use crate::lib::ai::{AIGoal, AI};
use crate::lib::point::{Direction, Point};
use crate::lib::screen::Screen;

use std::collections::{HashMap, HashSet};

use ncurses::*;

use downcast_rs::{impl_downcast, Downcast};

pub trait Entity: AI + Downcast {
    fn get_position(&self) -> Point;
    fn get_id(&self) -> u32;
    fn set_id(&mut self, id: u32);
}


macro_rules! impl_entity {
    ($t: ident) => {
        impl Entity for $t {
            fn get_position(&self) -> Point {
                self.pos.clone()
            }

            fn get_id(&self) -> u32 {
                self.id
            }

            fn set_id(&mut self, id: u32) {
                self.id = id;
            }
        }
    };
}

impl_downcast!(Renderable);

#[derive(Clone)]
pub struct Ant {
    pub pos: Point,
    pub id: u32,
    pub goal: AIGoal,
    pub dir: Direction,
    pub history: HashSet<Point>,
}

impl Ant {
    pub fn new(x: i32, y: i32) -> Ant {
        Ant {
            pos: Point(x, y),
            id: 0,
            goal: AIGoal::Seek,
            dir: Direction::Up,
            history: HashSet::new(),
        }
    }

    pub fn forward(&mut self, s: &Screen) {
        let target = self.dir.relative_point(&self.pos);
        if s.is_in_bounds(&target) {
            self.history.insert(self.pos);
            self.pos = target;
        } else {
            self.cw();
            self.cw();
        }
    }
    
    pub fn ccw(&mut self) {
        self.dir = self.dir.ccw();
    }

    pub fn cw(&mut self) {
        self.dir = self.dir.cw();
    }

    pub fn about_face(&mut self) {
        self.cw();
        self.cw();
    }
}

impl_entity!(Ant);
impl_entity!(Food);
impl_entity!(FoodGenerator);
impl_entity!(Egg);
impl_entity!(Queen);

impl Renderable for Ant {
    fn representation(&self) -> &str {
        match self.dir {
            Direction::Up => "^",
            Direction::Down => "v",
            Direction::Right => ">",
            Direction::Left => "<",
        }
    }
    fn before_render(&self) {
        attron(A_BOLD());
        match self.goal {
            AIGoal::Return => attron(A_UNDERLINE()),
            AIGoal::Seek => 0,
        };
    }

    fn after_render(&self) {
        attroff(A_BOLD());
        match self.goal {
            AIGoal::Return => attroff(A_UNDERLINE()),
            AIGoal::Seek => 0,
        };
    }
}

#[derive(Clone)]
pub struct Queen {
    pub pos: Point,
    pub egg_count: u8,
    pub id: u32,
}

impl Renderable for Queen {
    fn representation(&self) -> &str {
        "q"
    }

    fn before_render(&self) {
        attron(A_BOLD());
    }

    fn after_render(&self) {
        attroff(A_BOLD());
    }
}

pub trait Renderable: Entity {
    fn representation(&self) -> &str {
        "X"
    }
    fn before_render(&self) {}
    fn after_render(&self) {}
    fn render(&self, b: &Screen) {
        self.before_render();
        b.render(&self.get_position(), self.representation());
        self.after_render();
    }
}

impl Queen {
    pub fn new(x: i32, y: i32) -> Queen {
        Queen {
            pos: Point(x, y),
            egg_count: 0,
            id: 0,
        }
    }
}

#[derive(Debug, Clone)]
pub struct Egg {
    pub pos: Point,
    pub counter: u8,
    pub queen_id: u32,
    pub id: u32,
}

impl Renderable for Egg {
    fn representation(&self) -> &str {
        "e"
    }

    fn before_render(&self) {
        attron(A_BOLD());
    }

    fn after_render(&self) {
        attroff(A_BOLD());
    }
}

impl Egg {
    pub fn new(x: i32, y: i32, queen_id: u32) -> Egg {
        Egg {
            pos: Point(x, y),
            counter: 10,
            id: 0,
            queen_id,
        }
    }
}

pub struct Entities {
    pub data: HashMap<u32, Box<dyn Renderable>>,
    id_counter: u32,
}

impl Entities {
    pub fn new() -> Entities {
        Entities {
            id_counter: 0,
            data: HashMap::new(),
        }
    }

    pub fn add_entity<T: Renderable + Clone>(&mut self, e: &T) -> u32 {
        let mut clone = e.clone();
        clone.set_id(self.id_counter);
        let id = clone.get_id();
        self.data.insert(id, Box::new(clone));
        self.id_counter += 1;
        e.get_id()
    }
}

#[derive(Clone)]
pub struct Food {
    pub pos: Point,
    id: u32,
}

impl Food {
    pub fn new(x: i32, y: i32) -> Food {
        Food {
            pos: Point(x, y),
            id: 0,
        }
    }
}

impl Renderable for Food {
    fn representation(&self) -> &str {
        "f"
    }
}

// no position, does not get rendered yet acts per turn
#[derive(Clone)]
pub struct FoodGenerator {
    pub timer: u32,
    pos: Point,
    id: u32,
    pub counter: u32,
}

impl FoodGenerator {
    pub fn new() -> FoodGenerator {
        FoodGenerator {
            timer: 0,
            id: 0,
            pos: Point(0, 0),
            counter: 0,
        }
    }
}

impl Renderable for FoodGenerator {
    fn render(&self, b: &Screen) {}
}
